//
//  SPK_CalligraphyStrokesAnimator.m
//  SPK-Calligraphy
//
//  Created by Toby Harris on 25/03/2009.
//  Copyright (c) 2009 *spark audiovisual. All rights reserved.
//
//  'SPK-Calligraphy' is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  'SPK-Calligraphy' is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>

#import "SPK_CalligraphyStrokesAnimator.h"

#define	kQCPlugIn_Name				@"SPK Calligraphy - Strokes Animator"
#define	kQCPlugIn_Description		@"Use with Strokes Recorder. Designed for use displaying the strokes in handwritten sentences. If Animate is true, the strokes are drawn as they were recorded (framerate dependent). Experiment with flow values, use decay proportion of 1 and/or a minimum of 1 to disable."

@implementation SPK_CalligraphyStrokesAnimator

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

// @dynamic inputStructureOfLineStructures, inputColor, inputFlowDecay, inputFlowMinimum, inputWidth, inputOriginX, inputOriginY, inputPosZ, inputAnimate;
@dynamic inputFlowDecay, inputFlowMinimum, inputColor, inputWidth, inputAnimate, inputPosZ, inputOriginY, inputOriginX, inputStructureOfLineStructures;

+ (NSDictionary*) attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary*) attributesForPropertyPortWithKey:(NSString*)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	
	if([key isEqualToString:@"inputStructureOfLineStructures"])
		return [NSDictionary dictionaryWithObjectsAndKeys:@"Lines Structure", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputColor"])
		return [NSDictionary dictionaryWithObjectsAndKeys:@"Line Colour", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputWidth"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Line Width", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.05], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0.001], QCPortAttributeMinimumValueKey,
				nil];
	if([key isEqualToString:@"inputFlowDecay"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Flow Decay Proportion", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.9], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0], QCPortAttributeMinimumValueKey,
				[NSNumber numberWithDouble:1], QCPortAttributeMaximumValueKey,
				nil];
	if([key isEqualToString:@"inputFlowMinimum"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Flow Minimum Proportion", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0.2], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0], QCPortAttributeMinimumValueKey,
				[NSNumber numberWithDouble:1], QCPortAttributeMaximumValueKey,
				nil];
	if([key isEqualToString:@"inputOriginX"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"X Origin", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0], QCPortAttributeDefaultValueKey,
				nil];
	if([key isEqualToString:@"inputOriginY"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Y Origin", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0], QCPortAttributeDefaultValueKey,
				nil];
	if([key isEqualToString:@"inputPosZ"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Z Position", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:0], QCPortAttributeDefaultValueKey,
				nil];
	if([key isEqualToString:@"inputAnimate"])
		return [NSDictionary dictionaryWithObjectsAndKeys:@"Animate Lines", QCPortAttributeNameKey, nil];
	return nil;
}

+ (QCPlugInExecutionMode) executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	
	return kQCPlugInExecutionModeConsumer;
}

+ (QCPlugInTimeMode) timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeNone;
}

- (QCPlugInViewController*) createViewController
{
	/*
	Return a new QCPlugInViewController to edit the internal settings of this plug-in instance.
	You can return a subclass of QCPlugInViewController if necessary.
	*/
	
	return [[QCPlugInViewController alloc] initWithPlugIn:self viewNibName:@"Settings"];
}

@end

@implementation SPK_CalligraphyStrokesAnimator (Execution)

- (BOOL) startExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	CGLContextObj					cgl_ctx = [context CGLContextObj];
	GLenum							error;
	
	// Get index that our display list will be named by
	displayListIndex = glGenLists(1);
		
	// check for OpenGL errors
	if(error = glGetError())
		[context logMessage:@"OpenGL error %04X", error];
	
	return (error ? NO : YES);
}

- (void) enableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/

	// Reset animation to beginning
	if (self.inputAnimate) countRunning = 0;
}

- (BOOL) execute:(id<QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary*)arguments
{
	CGLContextObj					cgl_ctx = [context CGLContextObj];
	GLenum							error;
	GLint							blendEnabled;
	
	const CGFloat*					colorComponents;
	float							decay, baseAlpha, runningAlpha;
	NSArray*						linesArray;
	NSArray*						lineStructure;
	NSArray*						xypCoordinates;
	double							x, y, width, offset;
	double							originX, originY, zPos;
	NSInteger						count, countAtLineBeginning, countAtLineEnd, countTotal, i;
	BOOL							needRedraw;
	
	if(cgl_ctx == NULL)
		return NO;
	//// TASK: Determine need to redraw, animate etc.
	
	needRedraw = NO;
	
	// Get the total count of line points, by summing the count of each line structure in the lines array
	linesArray = self.inputStructureOfLineStructures;
	countPreviousTotal = countTotal;
	countTotal = [[linesArray valueForKeyPath:@"@sum.@count"] integerValue];
	
	// If we are animating...
	if (self.inputAnimate) {
		// If the total count is zero, reset the animation.
		if (countTotal == 0) {
			countRunning = 0;
		}
		
		// If the input is toggled on (as we're ANDed here)
		if ([self didValueForInputKeyChange:@"inputAnimate"]) {
			countRunning = 0;
		}
		
		// Are we still animating through the lines?
		if (countRunning <= countTotal) {
			needRedraw = YES;
		}
	}
	
	// Have the lines changed?
	if (countTotal != countPreviousTotal) {
		needRedraw = YES;
	}
	
	//// TASK: Set up blend mode to allow alpha
	blendEnabled = glIsEnabled(GL_BLEND);
	if(!blendEnabled)
		glEnable(GL_BLEND);
	glBlendEquation(GL_FUNC_ADD);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	//// TASK: Create new display list if required and send openGL commands
	
	if (needRedraw) {
		// cache inputs we're going to need repeatedly
		width = self.inputWidth;
		decay = self.inputFlowDecay;
		originX = self.inputOriginX;
		originY = self.inputOriginY;
		zPos = self.inputPosZ;
		
		// get RGBA components (the CGColorRef is guaranteed to be of type RGBA, no need to save / restore as the current color is part of the GL_CURRENT_BIT)
		colorComponents = CGColorGetComponents(self.inputColor);	
		
		// set our flow fade out vars
		baseAlpha = colorComponents[3] * self.inputFlowMinimum;
		
		countAtLineEnd = -1;
		
		glNewList(displayListIndex, GL_COMPILE_AND_EXECUTE);
		
		// Generate the line as a QUAD STRIP as if we had a chisel nib at 45°
		for (lineStructure in linesArray) {
			glBegin(GL_QUAD_STRIP);
			countAtLineBeginning = countAtLineEnd + 1;
			countAtLineEnd = countAtLineBeginning + [lineStructure count] - 1;
			runningAlpha = colorComponents[3] - baseAlpha;
			if (countRunning < countAtLineEnd && self.inputAnimate) {
				count = countRunning - countAtLineBeginning; // this can create negative number which stops further strokes from drawing
				for(i = 0; i < count; ++i) {
					xypCoordinates = [lineStructure objectAtIndex:i];
					x = [[xypCoordinates objectAtIndex:0] doubleValue] + originX;
					y = [[xypCoordinates objectAtIndex:1] doubleValue] + originY;
					offset = (width / 2) * ([[xypCoordinates objectAtIndex:2] doubleValue] + 0.5);
					runningAlpha *= decay;
					
					glColor4f(colorComponents[0], colorComponents[1], colorComponents[2], baseAlpha + runningAlpha);
					glVertex3d(x+offset, y+offset, zPos);
					glVertex3d(x-offset, y-offset, zPos);
				}
			} else {
				for (xypCoordinates in lineStructure) {
					x = [[xypCoordinates objectAtIndex:0] doubleValue] + originX;
					y = [[xypCoordinates objectAtIndex:1] doubleValue] + originY;
					offset = (width / 2) * ([[xypCoordinates objectAtIndex:2] doubleValue] + 0.5);
					runningAlpha *= decay;
					
					glColor4f(colorComponents[0], colorComponents[1], colorComponents[2], baseAlpha + runningAlpha);
					glVertex3d(x+offset, y+offset, zPos);
					glVertex3d(x-offset, y-offset, zPos);
				}
			}
			glEnd();
		}		
		glEndList();
		
		countRunning++;
	} else {
		glCallList(displayListIndex);	
	}
		
	if(!blendEnabled)
		glDisable(GL_BLEND);
	
	// check for OpenGL errors
	if(error = glGetError())
		[context logMessage:@"OpenGL error %04X", error];
	
	return (error ? NO : YES);
}

- (void) disableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/

}

- (void) stopExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
	
	CGLContextObj cgl_ctx = [context CGLContextObj];
	glDeleteLists(displayListIndex, 1);
}

@end
